"""
Utilities for waiting.
"""
import asyncio
import logging
import time
from typing import Callable

from openoligo.utils.sim import SIMULATION_SPEEDUP_FACTOR


async def wait_async(seconds: float) -> None:
    """Wait for a given number of seconds."""
    logging.debug("Start waiting for %.2f seconds", seconds)
    await asyncio.sleep(seconds / SIMULATION_SPEEDUP_FACTOR)
    logging.debug("Done waiting for %.2f seconds", seconds)


def wait(seconds: float) -> None:
    """Wait for a given number of seconds."""
    logging.debug("Start waiting for %.2f seconds", seconds)
    time.sleep(seconds / SIMULATION_SPEEDUP_FACTOR)
    logging.debug("Done waiting for %.2f seconds", seconds)


async def with_wait_async(func: Callable, seconds: float, *args, **kwargs) -> None:
    """Call a function, then wait for a given number of seconds."""
    func(*args, **kwargs)
    await wait_async(seconds / SIMULATION_SPEEDUP_FACTOR)


def with_wait(
    func: Callable, seconds: float, *args, **kwargs
) -> None:  # pylint: disable=unused-argument
    """Call a function, then wait for a given number of seconds."""
    func(*args, **kwargs)
    wait(seconds / SIMULATION_SPEEDUP_FACTOR)


def ms(seconds: float) -> float:  # pylint: disable=invalid-name
    """Convert seconds to milliseconds."""
    return seconds / 1000



"""
DNA Snthesis Protocol
"""
import logging
from time import time

from Bio.Seq import Seq  # type: ignore
from tqdm import tqdm

from openoligo import utils
from openoligo.instrument import Instrument
from openoligo.steps.flow import dry_all, send_to_waste_rxn, solvent_wash_all
from openoligo.steps.types import step
from openoligo.utils import wait_async


@step
async def detritylate(instrument: Instrument) -> None:
    """
    Add 3% trichloroacetic acid in dichloromethane to the reactor
    """
    await send_to_waste_rxn(instrument, "act")
    await wait_async(50)


@step
async def activate(instrument: Instrument) -> None:
    """
    Add 0.1 M phosphoramidite monomer and 0.5 M tetrazole to the reactor
    """
    await send_to_waste_rxn(instrument, "act")
    await wait_async(50)


@step
async def cap(instrument: Instrument) -> None:
    """
    Add acetic anhydride/pyridine/THF and N-methyl imidazole to the reactor
    """
    await send_to_waste_rxn(instrument, "cap1")
    await send_to_waste_rxn(instrument, "cap2")
    await wait_async(30)


@step
async def oxidize(instrument: Instrument) -> None:
    """
    Add 0.015 M iodine in water/pyridine/THF to the reactor
    """
    await send_to_waste_rxn(instrument, "oxi")
    await wait_async(45)


@step
async def cleave(instrument: Instrument) -> None:
    """
    Cleave the DNA sequence from the solid support.
    """
    await send_to_waste_rxn(instrument, "clde")
    await wait_async(180)


@step
async def deprotect(instrument: Instrument) -> None:
    """
    Remove the protecting groups from the DNA sequence.
    """
    await send_to_waste_rxn(instrument, "deb")
    await wait_async(45)


async def synthesize(instrument: Instrument, seq: Seq) -> None:
    """
    Synthesize a DNA sequence.

    args:
        seq: DNA sequence to synthesize.
    """
    logging.info("Initiating synthesis of DNA sequence: '%s'", seq)
    start_time = time()  # start timer
    with tqdm(total=len(seq) + 2) as pbar:
        await solvent_wash_all(instrument)
        await dry_all(instrument)

        pbar.update(1)

        for base_index, base in enumerate(seq):
            logging.info("Adding %sth base '%s' to growing DNA strand", base_index, base)

            await detritylate(instrument)

            await solvent_wash_all(instrument)
            await dry_all(instrument)

            await activate(instrument)

            await solvent_wash_all(instrument)
            await dry_all(instrument)

            await cap(instrument)

            await solvent_wash_all(instrument)
            await dry_all(instrument)

            await oxidize(instrument)

            await solvent_wash_all(instrument)
            await dry_all(instrument)

            await wait_async(1)
            pbar.update(1)

        await cleave(instrument)
        await deprotect(instrument)

        pbar.update(1)
    end_time = time()  # end timer
    elapsed_time_in_minutes = ((end_time - start_time) / 60) * utils.SIMULATION_SPEEDUP_FACTOR
    logging.info(
        "Synthesis complete for DNA sequence: '%s' in %s minutes", seq, elapsed_time_in_minutes
    )



"""
Logging configuration for OpenOligo
"""
import logging
import os
from logging.handlers import RotatingFileHandler

from dotenv import load_dotenv
from rich.logging import RichHandler

# Load the environment variables from the .env file
load_dotenv()

# Configure the logger
CONSOLE_FORMAT = "<%(filename)s:%(lineno)d> %(message)s"
FILE_FORMAT = "%(asctime)s <%(filename)s:%(lineno)d> [%(levelname)s] %(message)s"
# DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
# DATE_FORMAT = "%H:%M:%S"
LOG_LEVEL = os.getenv("OO_LOG_LEVEL", "INFO")

logger = logging.getLogger()
logger.setLevel(logging.NOTSET)

rot = RotatingFileHandler("openoligo.log", maxBytes=1000000, backupCount=5)
rot.setFormatter(logging.Formatter(FILE_FORMAT))
rot.setLevel(logging.DEBUG)

rich = RichHandler()
rich.setFormatter(logging.Formatter(CONSOLE_FORMAT))
rich.setLevel(logging.getLevelName(LOG_LEVEL))

logger.addHandler(rot)
logger.addHandler(rich)



"""
Sets some parameters for the simulation.
"""
import logging
import os

from openoligo.hal.types import __platform__, Platform

# Simulation parameters
# Can be set using env variable OO_SIM_SPEED=1000
# SIMULATION_SPEEDUP_FACTOR = 1 if is_rpi() else int(os.environ.get("OO_SIM_SPEED", 1000))
if os.environ.get("OO_SIM_SPEED") or __platform__ == Platform.SIM:
    SIMULATION_SPEEDUP_FACTOR = int(os.environ.get("OO_SIM_SPEED", 1000))
else:
    SIMULATION_SPEEDUP_FACTOR = 1

logging.info(
    "Using simulation speedup factor of %d (60min to %.2fs)",
    SIMULATION_SPEEDUP_FACTOR,
    60 * 60 / SIMULATION_SPEEDUP_FACTOR,
)



"""
Imports all the classes and functions from the openoligo package.
This is for convenience only. It is recommended to import only the
classes and functions that are needed.
"""

# flake8: noqa

from openoligo import log_config  # pylint: disable=unused-import
from openoligo.hal import *
from openoligo.instrument import *
from openoligo.utils import *

__all__ = [
    "Board",
    "Valve",
    "Switch",
    "ValveState",
    "ValveType",
]



"""
Imports all the modules in the utils directory.
"""
from openoligo.utils.sim import SIMULATION_SPEEDUP_FACTOR
from openoligo.utils.wait import ms, wait, wait_async, with_wait

__all__ = ["ms", "wait", "wait_async", "with_wait", "SIMULATION_SPEEDUP_FACTOR"]



"""
Pins in the Raspberry Pi GPIO header.
"""
from typing import Dict, TypedDict

from openoligo.hal.devices import Switch, Valve
from openoligo.hal.types import Board, NoSuchPinInPinout, Switchable, ValveRole
from openoligo.utils.singleton import Singleton


# Initialize fixed pinout
class FixedPinoutDict(TypedDict):
    """
    A dictionary of pinout that can not be configured by the library users.
    """

    waste: Valve
    waste_rxn: Valve
    prod: Valve
    branch: Valve
    rxn_out: Valve
    sol: Valve
    gas: Valve
    valve_pressure: Switch
    gas_pressure: Switch


fixed_pinout: FixedPinoutDict = {
    "waste": Valve(gpio_pin=Board.P7, role=ValveRole.OUTLET),
    "waste_rxn": Valve(gpio_pin=Board.P5, role=ValveRole.OUTLET),
    "prod": Valve(gpio_pin=Board.P8, role=ValveRole.OUTLET),
    "branch": Valve(gpio_pin=Board.P12, role=ValveRole.BRANCH),
    "rxn_out": Valve(gpio_pin=Board.P13, role=ValveRole.TRANSIT),
    "sol": Valve(gpio_pin=Board.P3),
    "gas": Valve(gpio_pin=Board.P10),
    "valve_pressure": Switch(gpio_pin=Board.P11),
    "gas_pressure": Switch(gpio_pin=Board.P29),
}


def list_configurable_pins() -> dict[str, Board]:
    """
    Returns a list of all configurable pins.
    These ar all pins in the board that are not part of the fiixed pinout.
    """
    fixed_pins = [sw.gpio_pin for sw in fixed_pinout.values()]  # type: ignore
    return [pin for pin in Board if pin not in fixed_pins]  # type: ignore


class Pinout(metaclass=Singleton):
    """
    Pinout for the instrument.
    """

    def __init__(
        self,
        phosphoramidites: Dict[str, Valve],
        reactants: Dict[str, Valve],
    ):
        """
        Initialize the pinout.
        """
        self.fixed = fixed_pinout
        self.phosphoramidites = phosphoramidites
        self.reactants = reactants

        self.__pinout: Dict[str, Switchable] = {}
        self.init_pinout()

    def __add_pinout_safe(self, pin: Switchable, category: str, sub_category: str):
        """
        Check for duplicate pin usage.

        args:
            pin: Pin to check if its already in the pinout.
            category: Category of the pin. (Just for logging)
            sub_category: Sub category of the pin. (the name of the pin)
        """
        if pin in self.__pinout.values():
            raise ValueError(f"Pin {pin} is used twice in {category}.{sub_category}.")
        self.__pinout[sub_category.lower()] = pin

    def init_pinout(self):
        """
        Make sure no pin is used twice.
        """
        for key, switchables in self.__dict__.items():
            if key.startswith("_"):
                continue

            if isinstance(switchables, dict):
                for category, pin in switchables.items():
                    self.__add_pinout_safe(pin, key, category)
            else:
                self.__add_pinout_safe(switchables, self.__class__.__name__, key)

    def pins(self) -> dict[str, Switchable]:
        """
        Return a list of all pins in the pinout.
        """
        return self.__pinout

    def valves(self) -> dict[str, Valve]:
        """
        Return a list of all valves in the pinout.
        """
        return {k: v for k, v in self.__pinout.items() if isinstance(v, Valve)}

    def get(self, name: str) -> Switchable:
        """
        Return the switch/valve with the given name.

        args:
            name: Name of the pin.

        returns:
            The switch/valve with the given name.

        raises:
            KeyError: If the pin is not found.
        """
        try:
            return self.__pinout[name.lower()]
        except KeyError as exc:
            raise NoSuchPinInPinout(
                f"Pin {name} not found in pinout. \n\nAvailable pins are: {list(self.__pinout)}"
            ) from exc

    def __repr__(self):
        """
        Return a string representation of the pinout.
        """
        return f"Pinout({self.__pinout})"



"""
This module contains definitions of protocols and exceptions related to switchable devices.
"""
from enum import Enum
from typing import Protocol


def is_rpi() -> bool:
    """Returns True if running on a Raspberry Pi, False otherwise."""
    try:
        with open("/proc/cpuinfo", "r", encoding="utf-8") as file:
            return "Raspberry" in file.read()
    except FileNotFoundError:
        return False


def is_bb() -> bool:
    """Returns True if running on a Raspberry Pi, False otherwise."""
    try:
        with open("/proc/cpuinfo", "r", encoding="utf-8") as file:
            return "AM33XX" in file.read()
    except FileNotFoundError:
        return False


class Platform(Enum):
    """Platform type."""

    RPI = "RPI"
    BB = "BB"
    SIM = "SIM"


def get_platform() -> Platform:
    """Get the platform type."""
    if is_rpi():
        return Platform.RPI
    if is_bb():
        return Platform.BB
    return Platform.SIM


class RPiBoard(Enum):
    """
    Enumerates the GPIO pins on the Raspberry Pi 3 Model B.
    """

    P3 = 3
    P5 = 5
    P7 = 7
    P8 = 8
    P10 = 10
    P11 = 11
    P12 = 12
    P13 = 13
    P15 = 15
    P16 = 16
    P18 = 18
    P19 = 19
    P21 = 21
    P22 = 22
    P23 = 23
    P24 = 24
    P26 = 26
    P27 = 27
    P28 = 28
    P29 = 29
    P31 = 31
    P32 = 32
    P33 = 33
    P35 = 35
    P36 = 36
    P37 = 37
    P38 = 38
    P40 = 40


class BbBoard(Enum):
    """
    Enumerates the GPIO pins on the Beaglebone Black.
    """

    P3 = 3
    P5 = 5
    P7 = 7
    P8 = 8
    P10 = 10
    P11 = 11
    P12 = 12
    P13 = 13
    P15 = 15
    P16 = 16
    P18 = 18
    P19 = 19
    P21 = 21
    P22 = 22
    P23 = 23
    P24 = 24
    P26 = 26
    P27 = 27
    P28 = 28
    P29 = 29
    P31 = 31
    P32 = 32
    P33 = 33
    P35 = 35
    P36 = 36
    P37 = 37
    P38 = 38
    P40 = 40


__platform__ = get_platform()


if __platform__ == Platform.BB:
    Board = BbBoard
else:
    Board = RPiBoard  # Also acts as a mock board


class GpioMode(Enum):
    """GPIO mode."""

    IN = 0
    OUT = 1


class ValveState(Enum):
    """This class represents the state of a valve."""

    OPEN_FLOW = 0
    CLOSED_FLOW = 1


class ValveType(Enum):
    """Valve Type: NC/NO"""

    NORMALLY_CLOSED = 0
    NORMALLY_OPEN = 1


class ValveRole(Enum):
    """Valve Role: Inlet/Outlet/Transit/Branch"""

    INLET = 0
    OUTLET = 1
    TRANSIT = 2
    BRANCH = 3


class Switchable(Protocol):
    """
    This protocol represents a switchable device. Any class implementing this protocol should
    be able to set, get the current value and toggle the state of the switch.
    """

    def __init__(self, pin: int, name: str):
        """Initialize the switchable device."""
        raise NotImplementedError

    def set(self, state: bool):
        """Set the state of the switch."""
        raise NotImplementedError

    @property
    def gpio(self) -> Board:
        """Get the GPIO pin number."""
        raise NotImplementedError

    @property
    def value(self) -> bool:
        """Get the current value of the switch."""
        raise NotImplementedError


class Valvable(Switchable, Protocol):
    """
    A semantic wrapper around a switchable device. This class is used to represent a valve.
    """

    def __init__(self, pin: int, name: str, valve_type: ValveType = ValveType.NORMALLY_OPEN):
        """Initialize the valve."""
        raise NotImplementedError

    def open(self):
        """Open the valve."""
        raise NotImplementedError

    def close(self):
        """Close the valve."""
        raise NotImplementedError

    @property
    def get_type(self) -> ValveType:
        """Get the type of the valve."""
        raise NotImplementedError


class SwitchingError(Exception):
    """
    An exception that represents a failure in switching operation.
    """


class InvalidManifoldSizeError(Exception):
    """
    An exception that represents an invalid manifold size.
    """


class NoSuchPinInPinout(Exception):
    """
    Thrown when a pin for a given name is not found in the pinout.
    """


class OneSourceException(SwitchingError):
    """
    When more than one source valve is being attempted to be used.
    """


class OneDestinationException(SwitchingError):
    """
    When more than one destination valve is being attempted to be used.
    """



"""
GPIO access abstraction: RPi.GPIO on Raspberry Pi, MockGPIO otherwise.
"""
import importlib
import logging
from abc import ABC, abstractmethod

from openoligo.hal.types import Board, GpioMode, is_rpi


def get_gpio():
    """Get the GPIO module."""
    try:
        with open("/proc/cpuinfo", "r", encoding="utf-8") as file:
            if "Raspberry" in file.read():
                gpio = importlib.import_module("RPi.GPIO")
                return RPiGPIO(gpio)
            return MockGPIO()
    except (FileNotFoundError, ModuleNotFoundError):
        return MockGPIO()


class GPIOInterface(ABC):
    """GPIO interface."""

    @abstractmethod
    def setup(self) -> None:
        """Setup GPIO."""

    @abstractmethod
    def setup_pin(self, pin: Board, mode: GpioMode) -> None:
        """Set up a GPIO pin."""

    @abstractmethod
    def set(self, pin: Board, value: bool) -> None:
        """Set the output value of a GPIO pin."""
        logging.debug("Setting pin %s to %s", pin, value)

    @abstractmethod
    def value(self, pin: Board) -> bool:
        """Get the value of a GPIO pin."""

    @abstractmethod
    def cleanup(self) -> None:
        """Clean up GPIO."""

    @abstractmethod
    def __repr__(self) -> str:
        """Return a string representation of the GPIO module."""


class RPiGPIO(GPIOInterface):
    """GPIO access on Raspberry Pi."""

    def __init__(self, gpio):
        """Import Board.GPIO"""
        self.gpio = gpio

    def setup(self):
        """Set up all pins."""
        for pin in Board:
            self.setup_pin(pin)

    def setup_pin(self, pin: Board, mode: GpioMode = GpioMode.OUT) -> None:
        """Set up a GPIO pin."""
        self.gpio.setmode(self.gpio.BOARD)
        self.gpio.setup(pin.value, self.gpio.OUT if mode == GpioMode.OUT else self.gpio.IN)

    def set(self, pin: Board, value: bool) -> None:
        """Set the output value of a GPIO pin."""
        super().set(pin, value)
        self.gpio.output(pin.value, self.gpio.HIGH if value else self.gpio.LOW)

    def value(self, pin: Board) -> bool:
        return self.gpio.input(pin.value)

    def cleanup(self):
        """Clean up GPIO."""
        self.gpio.cleanup()

    def __repr__(self) -> str:
        """Return a string representation of the GPIO module."""
        return ", ".join(f"({pin.value}, {1 if self.value(pin) else 0})" for pin in Board)


class MockGPIO(GPIOInterface):
    """Mock GPIO access."""

    def __init__(self):
        """
        Initialize the state. Its keys are the pin numbers,
        and the values are their mock states
        """
        self.state: dict[int, bool] = {}

    def setup(self) -> None:
        """Setup all pins."""
        for pin in Board:
            self.setup_pin(pin)

    def setup_pin(
        self, pin: Board, mode: GpioMode = GpioMode.OUT
    ) -> None:  # pylint: disable=unused-argument
        """Set up a GPIO pin."""
        self.state[pin.value] = False

    def set(self, pin: Board, value: bool):
        """Set the output value of a GPIO pin."""
        super().set(pin, value)
        self.state[pin.value] = value

    def value(self, pin: Board) -> bool:
        """Get the value of a GPIO pin."""
        return self.state[pin.value]

    def cleanup(self):
        """Clean up GPIO."""
        for pin in Board:
            self.set(pin, False)

    def __repr__(self) -> str:
        """Return a string representation of the GPIO module."""
        return ", ".join(f"({pin.value}, {1 if self.value(pin) else 0})" for pin in Board)



"""
Functions to create various flows from reagents to columns,
and to clean columns.
"""
import logging

from openoligo import Instrument
from openoligo.steps.types import FlowBranch, substep
from openoligo.utils.wait import wait_async


@substep
async def send_to_prod(instrument: Instrument, src: str) -> None:
    """
    Flow a reagent to the final product outlet.

    args:
        src: Slot to flow reagents from.
    """
    instrument.all_except([src, "prod", "branch", "rxn_out"])
    logging.debug("Flowing %s to prod", src)


@substep
async def send_to_waste_rxn(instrument: Instrument, src: str) -> None:
    """
    Flow a reagent to the final product outlet.

    args:
        src: Slot to flow reagents from.
    """
    instrument.all_except([src, "branch", "rxn_out", "waste_rxn"])
    logging.debug("Flowing %s to reaction waste", src)


@substep
async def solvent_wash(instrument: Instrument, branch: FlowBranch, duration: float = 10) -> None:
    """
    Wash column with solvent.

    args:
        branch: FlowBranch to wash.
    """
    logging.debug("Initiating washing flow branch %s with solvent", branch)

    if branch == FlowBranch.REACTION:
        instrument.all_except(["sol", "rxn_out", "branch", "waste_rxn"])
    elif branch == FlowBranch.REAGENTS:
        instrument.all_except(["sol", "waste"])

    await wait_async(duration)

    logging.debug("Washing of branch %s complete", branch)


@substep
async def solvent_wash_all(instrument: Instrument) -> None:
    """
    Wash all flow branches with solvent.
    """

    await solvent_wash(instrument, FlowBranch.REACTION)
    await solvent_wash(instrument, FlowBranch.REAGENTS)


@substep
async def dry(instrument: Instrument, branch: FlowBranch) -> None:
    """
    Dry column.

    args:
        branch: FlowBranch to dry.
    """
    logging.debug("Initiating drying of branch %s", branch)
    if branch == FlowBranch.REACTION:
        instrument.all_except(["gas", "rxn_out", "branch", "waste_rxn"])
    elif branch == FlowBranch.REAGENTS:
        instrument.all_except(["gas", "waste"])
    logging.debug("Drying of branch %s complete", branch)


@substep
async def dry_all(instrument: Instrument) -> None:
    """
    Dry all flow branches.
    """

    await dry(instrument, FlowBranch.REACTION)
    await dry(instrument, FlowBranch.REAGENTS)



"""
Types for steps
"""
import logging
from enum import Enum
from functools import wraps

FlowWaitPair = tuple[int, float]
FlowWaitPairs = list[FlowWaitPair]


class FlowBranch(Enum):
    """
    Flow branch enumeration.
    """

    REACTION = "reaction"
    REAGENTS = "reagents"


def _step_decorator(coroutine, is_substep: bool = False):
    """
    Decorator for a step function (which carries out a reaction step).
    """

    @wraps(coroutine)
    async def wrapper_coroutine(*args, **kwargs):
        # instrument = Instrument()

        name, doc = coroutine.__name__, coroutine.__doc__.strip().split("\n")[0]
        level = logging.DEBUG if is_substep else logging.INFO
        repr_args = args[1:]  # Remove instrument from args
        logging.log(
            level, "Starting step [bold]%s[/]%s: %s", name, repr_args, doc, extra={"markup": True}
        )

        return await coroutine(*args, **kwargs)

    return wrapper_coroutine


def step(coroutine):
    """
    Decorator for a step function (which carries out a reaction step).
    """
    return _step_decorator(coroutine, is_substep=False)


def substep(coroutine):
    """
    Decorator for a substep function (which carries out a sub-reaction step).
    """
    return _step_decorator(coroutine, is_substep=True)



"""
Unified access to configuring and using the instrument.
"""
import logging

from openoligo.hal.board import Pinout
from openoligo.hal.devices import Valve
from openoligo.hal.gpio import get_gpio
from openoligo.hal.types import Board, OneDestinationException, OneSourceException, ValveRole
from openoligo.utils.singleton import Singleton

default_pinout = Pinout(
    phosphoramidites={
        "A": Valve(gpio_pin=Board.P26),
        "C": Valve(gpio_pin=Board.P28),
        "G": Valve(gpio_pin=Board.P15),
        "T": Valve(gpio_pin=Board.P16),
    },
    reactants={
        "ACT": Valve(gpio_pin=Board.P18),
        "OXI": Valve(gpio_pin=Board.P19),
        "CAP1": Valve(gpio_pin=Board.P21),
        "CAP2": Valve(gpio_pin=Board.P22),
        "DEB": Valve(gpio_pin=Board.P23),
        "CLDE": Valve(gpio_pin=Board.P24),
    },
)


class Instrument(metaclass=Singleton):
    """
    Unified access to the instrument. Hide the details of the hardware.
    """

    def __init__(self, pinout: Pinout = default_pinout) -> None:
        """
        Initialize the instrument.
        """
        self.pinout = pinout

        logging.info("Initializing instrument with pinout: %s", self.pinout)

        self.__setup()

    def __setup(self):
        """
        Setup the instrument.
        """
        self.controller = get_gpio()
        self.controller.setup()

    def __get_valve(self, name: str) -> Valve:
        """
        Get a valve by name.
        """
        valve = self.pinout.get(name)
        if not isinstance(valve, Valve):
            raise ValueError(f"{name} is not a valve")
        return valve

    def validate_valve_set(self, name: list[str]) -> None:
        """
        Validate that the list of valves conform to a set of rules.

        1. Each valve must be a valve.
        2. The list must contain one and only one reagent (inlet) valve.
        3. The list must contain one and only one product (outlet) valve.
        4. The list may contain zero or more transit valves
        5. The list may contain a branch valve, but only one.

        args:
            name: list of valve names

        raises:
            OneSourceException: If there is not exactly one source valve.
            OneDestinationException: If there is not exactly one destination valve.
        """
        valve_types: dict[ValveRole, list[Valve]] = {
            ValveRole.INLET: [],
            ValveRole.OUTLET: [],
            ValveRole.BRANCH: [],
            ValveRole.TRANSIT: [],
        }

        for _name in name:
            if not isinstance(self.pinout.get(_name), Valve):
                raise ValueError(f"{_name} is not a valve")
            valve: Valve = self.__get_valve(_name)

            if valve.role not in valve_types:
                raise ValueError(f"Unknown valve role: {valve.role}")

            valve_types[valve.role].append(valve)

        if len(valve_types[ValveRole.INLET]) != 1:
            raise OneSourceException("There must be exactly one source valve")

        if len(valve_types[ValveRole.OUTLET]) != 1:
            raise OneDestinationException("There must be exactly one destination valve")

        if len(valve_types[ValveRole.TRANSIT]) > 0:
            logging.debug("Make sure that the transit valve(s) are in the route you expect")

    def all_except(self, name: list[str]) -> None:
        """
        Set the state of a pin on the instrument.

        Args:
            name: A list of valve names
            state: The state of the pin.

        raises:
            OneSourceException: If there is not exactly one source valve.
            OneDestinationException: If there is not exactly one destination valve.
        """
        self.validate_valve_set(name)

        all_valves_in_pinout = self.pinout.valves()
        for _name in all_valves_in_pinout:
            if _name in name:
                all_valves_in_pinout[_name].open()
            else:
                all_valves_in_pinout[_name].close()
        logging.debug(
            "Set %s to [bold]open[/] and all others to close.", name, extra={"markup": True}
        )

    def __repr__(self):
        """
        Return a string representation of the instrument.
        """
        return f"Instrument({self.pinout})"



"""
Driver imports for convenience.
"""

from openoligo.hal.board import Pinout
from openoligo.hal.devices import Switch, Valve
from openoligo.hal.gpio import is_rpi
from openoligo.hal.types import (
    Board,
    GpioMode,
    InvalidManifoldSizeError,
    Switchable,
    SwitchingError,
    Valvable,
    ValveState,
    ValveType,
)

__all__ = [
    "is_rpi",
    "Valve",
    "Switch",
    "Board",
    "GpioMode",
    "Pinout",
    "InvalidManifoldSizeError",
    "Switchable",
    "SwitchingError",
    "Valvable",
    "ValveState",
    "ValveType",
]



"""
Switches can be used to control devices that can be turned on and off.
"""
import logging
from dataclasses import dataclass, field

from openoligo.hal.gpio import get_gpio
from openoligo.hal.types import Board, Switchable, Valvable, ValveRole, ValveState, ValveType


@dataclass
class Switch(Switchable):
    """
    A switch that actually controls a GPIO pin on the Raspberry Pi.

    Attributes:
        name: Name of the switch.
        gpio_pin: GPIO pin number.
        board: Board object to access the GPIO pins.
    """

    gpio_pin: Board
    _state: bool = field(default=False, init=False)
    _switch_count: int = field(default=0, init=False, repr=False)

    def __post_init__(self):
        """Initialize the switch."""
        self.controller = get_gpio()
        self.controller.setup_pin(self.gpio_pin)

    def set(self, state: bool):
        """Set state of the switch ON or OFF."""
        __old_state = self._state
        __new_state = state

        if __old_state == __new_state:
            logging.warning("Switch %s is already %s", self.gpio_pin, __old_state)
            return

        self._state = __new_state
        self.controller.set(self.gpio_pin, state)
        self._switch_count += 1

        logging.debug(
            "Switch (%s) set from %s to [bold]%s[/]",
            self.gpio_pin,
            __old_state,
            __new_state,
            extra={"markup": True},
        )

    @property
    def gpio(self) -> Board:
        """Get the GPIO pin number."""
        return self.gpio_pin

    def toggle(self):
        """Toggle the state of the switch."""
        self.set(not self._state)

    @property
    def value(self) -> bool:
        """
        Get the current value of the switch.
        """
        return self._state


@dataclass
class Valve(Valvable):
    """
    A valve that actually controls a GPIO pin on the Raspberry Pi.
    """

    gpio_pin: Board
    role: ValveRole = field(default=ValveRole.INLET)
    valve_type: ValveType = field(default=ValveType.NORMALLY_OPEN)
    _switch_count: int = field(init=False, default=0)
    _state: ValveState = field(init=False)

    def __post_init__(self):
        """Initialize the valve."""
        self.controller = get_gpio()
        self.controller.setup_pin(self.gpio_pin)

        self._state = (
            ValveState.CLOSED_FLOW
            if self.valve_type == ValveType.NORMALLY_CLOSED
            else ValveState.OPEN_FLOW
        )

    def open(self):
        """Open the valve."""
        self.set(True)

    def close(self):
        """Close the valve."""
        self.set(False)

    def set(self, state: bool):
        """Set the state of the valve."""
        __old_state: ValveState = self._state
        __new_state: ValveState = ValveState.OPEN_FLOW if state else ValveState.CLOSED_FLOW

        if __old_state == __new_state:
            if __old_state == ValveState.OPEN_FLOW:
                logging.debug("Valve (%s) is already %s", self.gpio_pin, __old_state)
            return

        self.controller.set(self.gpio_pin, state)
        self._state = __new_state

        self._switch_count += 1
        logging.debug(
            "Valve (%s) set from %s to [bold]%s[/]",
            self.gpio_pin,
            __old_state,
            self._state,
            extra={"markup": True},
        )

    @property
    def gpio(self) -> Board:
        """Get the GPIO pin number."""
        return self.gpio_pin

    @property
    def value(self) -> bool:
        """Get the current value of the valve."""
        return self._state == ValveState.OPEN_FLOW

    @property
    def get_type(self) -> ValveType:
        """Get the type of the valve."""
        return self.valve_type

    def __repr__(self) -> str:
        return f"{self.gpio_pin}[{self.value}]"



"""
Provides a Metaclass for creating Singleton classes.
"""
from typing import Any, Dict


class Singleton(type):
    """
    Metaclass for singleton classes.
    """

    _instances: Dict[type, Any] = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]
